#include "ShaderConstants.fxh"
#include "util.fxh"
#include "config.fxh"
#include "noise.fxh"

struct PS_Input
{
	float4 position : SV_Position;

#ifndef BYPASS_PIXEL_SHADER
	lpfloat4 color : COLOR;
	snorm float2 uv0 : TEXCOORD_0_FB_MSAA;
	snorm float2 uv1 : TEXCOORD_1_FB_MSAA;
	float3 cPos : chunkedPos;
	float3 wPos : worldPos;
	float isWater : is_Water;
#endif

#ifdef FOG
	float4 fogColor : FOG_COLOR;
#endif
};

struct PS_Output
{
	float4 color : SV_Target;
};

// Based on Filmic Tonemapping Operators http://filmicgames.com/archives/75
float3 tonemapFilmic(const float3 color) {
	float3 x = max(float3(0.0, 0.0, 0.0), color - 0.004);
	return (x * (6.2 * x + 0.5)) / (x * (6.2 * x + 1.7) + 0.06);
}

ROOT_SIGNATURE
void main(in PS_Input PSInput, out PS_Output PSOutput)
{
#ifdef BYPASS_PIXEL_SHADER
    PSOutput.color = float4(0.0f, 0.0f, 0.0f, 0.0f);
    return;
#else

#if USE_TEXEL_AA
	float4 diffuse = texture2D_AA(TEXTURE_0, TextureSampler0, PSInput.uv0 );
#else
	float4 diffuse = TEXTURE_0.Sample(TextureSampler0, PSInput.uv0);
#endif

#ifdef SEASONS_FAR
	diffuse.a = 1.0f;
#endif

#if USE_ALPHA_TEST
	#ifdef ALPHA_TO_COVERAGE
		#define ALPHA_THRESHOLD 0.05
	#else
		#define ALPHA_THRESHOLD 0.5
	#endif
	if(diffuse.a < ALPHA_THRESHOLD)
		discard;
#endif

#if defined(BLEND)
	diffuse.a *= PSInput.color.a;
#endif

#if !defined(ALWAYS_LIT)
	diffuse = diffuse * TEXTURE_1.Sample(TextureSampler1, saturate((PSInput.uv1 + float2(0.0, 0.1)) * float2(0.5, 2.0)));
#endif

#ifndef SEASONS
	#if !USE_ALPHA_TEST && !defined(BLEND)
		diffuse.a = PSInput.color.a;
	#endif	

	diffuse.rgb *= PSInput.color.rgb;
#else
	float2 uv = PSInput.color.xy;
	diffuse.rgb *= lerp(1.0f, TEXTURE_2.Sample(TextureSampler2, uv).rgb*2.0f, PSInput.color.b);
	diffuse.rgb *= PSInput.color.aaa;
	diffuse.a = 1.0f;
#endif

float indoor = smoothstep(1.0, 0.5, PSInput.uv1.y);
float daylight = TEXTURE_1.Sample(TextureSampler1, float2(0, 1)).r;
float sunset_sunrise = saturate(( FOG_COLOR.r - 0.1) - FOG_COLOR.b);
float rainStrength = smoothstep(0.5, 0.4, FOG_CONTROL.x);
float underwater;

if(FOG_CONTROL.x <= 0.2 && FOG_COLOR.b > FOG_COLOR.r + FOG_COLOR.g) {
	underwater = 1.0;
} else {
	underwater = 0.0;
}

float3 normal = normalize(cross(ddx(PSInput.cPos),ddy(PSInput.cPos)));

#ifdef ENABLE_TONEMAPPING
	// Increase the brightness of the world colour using tonemap function.
	// Colours used the traditional Japanese one.
	// See: https://www.colordic.org/w
	float3 dayColor = float3(1.53, 1.39, 1.42);				// 薄桜 Usu-zakura
	float3 nightColor = float3(0.61, 0.79, 0.82);			// 百入茶 Momoshio-cha
	float3 rainColor = float3(0.62, 0.85, 0.98);			// 褐返 Kachi-kaeshi
	float3 sunset_sunriseColor = float3(2.26, 0.4, 0.27);	// 猩々緋 Shoujouhi
	float3 indoorColor = float3(1.32, 1.11, 1.09);			// 灰桜 Hai-zakura
	float3 underwaterColor = float3(0, 0.58, 0.85);			// 青 Ao

	float3 worldColor = lerp(lerp(lerp(lerp(lerp(nightColor, dayColor, daylight), sunset_sunriseColor, sunset_sunrise), rainColor, rainStrength), indoorColor, indoor), underwaterColor, underwater);

	diffuse.rgb = tonemapFilmic(diffuse.rgb *(diffuse.rgb * worldColor));

#endif

#ifdef ENABLE_WATER_REFLECTION
    // Generate reflections of the sun, moon, and clouds on the water surface and underwater.
    // The game doesn't tell us the correct information, so these reflections are all fakes. 
	if(PSInput.isWater > 0.5 || bool(underwater == true)){
		float3 p = PSInput.cPos;
		float3 l = normalize(PSInput.wPos);
		p.xz *= float2(2.0, 0.5);

		float t = TOTAL_REAL_WORLD_TIME;

		float wave = snoise(t * 0.8 + p.xz) + snoise(t + float2(t * 0.51 + p.x, t * 0.2 + p.z));

		float4 refColor = lerp(float4(1, 1, 1, 1), float4(1, 0.5, 0.2, 1), sunset_sunrise);

		diffuse = lerp(lerp(diffuse * lerp(float4(0.9, 0.9, 0.9, 0.8), float4(1, 1, 1, 1), wave * 0.3), float4(diffuse.rgb * (1.0 + abs(pow(wave, 3.0)) * clamp(PSInput.uv1.y, 0.1, 0.8)), diffuse.a), underwater), refColor, lerp(lerp(lerp(lerp(lerp(0.0, lerp(0.0, saturate(noise((PSInput.wPos.xz+(1.0 - abs(wave)) * 6.0 * PSInput.wPos.xz / length(PSInput.wPos.xz)) * 0.15)), daylight), saturate(length(l.xz * 0.1 / l.y))), pow(abs(wave), 2.0), saturate(1.0 - length(l.z * lerp(4.0, 2.0, daylight))) * saturate(length(l.xz * 0.2 / l.y))), 0.0, underwater), 0.0, rainStrength), 0.0, indoor));
	}
#endif

#ifdef ENABLE_SHADOWS
	// Generate the shadows under the blocks.
	// It is generating on a dark point of the vanilla lightmap.
	// Game seems doesn't move the position of the lightmap, presumably due to a performance reason.
	// Therefore, we can't generate moveable shadows currently.
	diffuse.rgb *= lerp(1.0, lerp(0.5, 1.0, pow(PSInput.uv1.x, 5.0)), smoothstep(0.87, 0.865, PSInput.uv1.y));
#endif

#ifdef ENABLE_TORCHLIGHT
	// Generate the torchlight in the dark place.
	// Due to the effects of underwater visual effects, the underwater torchlight becomes another colour.
	// Colours used the traditional Japanese one.
	// See: https://www.colordic.org/w
	float3 normalLight = float3(0.93, 0.47, 0); 	// 橙色 daidai-iro
	float3 underwaterLight = float3(.17,.66,.88);	// 天色 Ama-iro

	float3 lightColor = lerp(normalLight, underwaterLight, underwater);

	float lightStrength = smoothstep(pow(daylight, 5.0) * (1.0 - rainStrength) * pow(PSInput.uv1.y, 5.0), 1.0, pow(PSInput.uv1.x, 3.0));

	diffuse.rgb += lightColor * lightStrength;

#endif

#ifdef ENABLE_FOG
	// Generate the fog around the player.
	// Its colour is as same as the vanilla fog colour.
	// The vanilla fog is generated when the chunk has updated, but our one is generated by the distance of your config.
	diffuse.rgb = lerp(diffuse.rgb, FOG_COLOR.rgb, smoothstep(0.0, RENDER_DISTANCE * 0.7, length(PSInput.wPos)));
#endif

#ifdef FOG
	diffuse.rgb = lerp( diffuse.rgb, PSInput.fogColor.rgb, PSInput.fogColor.a );
#endif

	PSOutput.color = diffuse;

#ifdef VR_MODE
	// On Rift, the transition from 0 brightness to the lowest 8 bit value is abrupt, so clamp to 
	// the lowest 8 bit value.
	PSOutput.color = max(PSOutput.color, 1 / 255.0f);
#endif

#endif // BYPASS_PIXEL_SHADER
}